@@ -5,6 +5,7 @@ communication with the client.
 
 from uuid import uuid1,UUID
 import trio
+from collections.abc import Mapping
 from typing import Optional,Dict,List,Union,Any
 from .codec import pack, unpack
 from functools import partial
@@ -15,6 +16,13 @@ processing = ContextVar("processing", default=None)
 import logging
 logger = logging.getLogger(__name__)
 
+async def _spawn(task, args=(), kwargs={}, task_status=trio.TASK_STATUS_IGNORED):
+    processing.set(None)
+
+    with trio.CancelScope() as sc:
+        task_status.started(sc)
+        await task(*args, **kwargs)
+
 
 class UnknownActionError(RuntimeError):
     """
@@ -48,8 +56,6 @@ class Talker:
 
     def __init__(self, websocket):
         self.ws = websocket
-        self.n = 1
-        self.req = {}
 
         global _talk_id
         self._id = _talk_id
@@ -87,7 +93,7 @@ class Talker:
             s.shield = True
             try:
                 if self.w:
-                    await self._send("fatal",msg)
+                    await self._send(["info",["danger",msg]])
             except Exception:
                 logger.exception("Terminal message")
                 pass
@@ -115,28 +121,13 @@ class Talker:
             await self.w.wait()
         while True:
             data = await self.ws.receive()
-            data = unpack(data)
+            try:
+                data = unpack(data)
+            except TypeError:
+                logger.error("IN X %r",data)
+                raise
             logger.debug("IN %r",data)
-            action,data = data
-            if action == "reply":
-                self._reply(*data)
-                continue
-
-            try:
-                res = getattr(self.w, 'msg_'+action)
-            except AttributeError:
-                res = partial(self.w.any_msg,action)
-            tk = processing.set((action,data))
-            try:
-                await res(data)
-            finally:
-                processing.reset(tk)
-
-    async def _reply(self, n, data):
-        if isinstance(data,Mapping) and '_error' in data:
-            data = ClientError(**data)
-        evt,self.req[n] = self.req[n],data
-        evt.set()
+            await self.w.data_in(data)
 
     async def ws_out(self, *, task_status=trio.TASK_STATUS_IGNORED):
         """
@@ -147,62 +138,31 @@ class Talker:
         if isinstance(self.w, trio.Event):
             await self.w.wait()
         while True:
-            action,data = await send_q.receive()
-            await self._send(action,data)
+            data = await send_q.receive()
+            await self._send(data)
 
-    async def _send(self, action, data):
-        data = [action,data]
+    async def _send(self, data):
         logger.debug("OUT %r",data)
         data = pack(data)
         await self.ws.send(data)
 
-    async def send(self, action,data):
+    async def send(self, data:Any):
         """
         Send a message to the client.
         """
-        if action == "req":
-            raise RuntimeError("Use '.request' for that!")
-        await self._send_q.send((action,data))
-
-    async def request(self, action,data):
-        """
-        Send a request to the client, await+return the reply
-        """
-        if processing.get():
-            raise RuntimeError("You cannot call this from within the receiver",processing.get())
-
-        self.n += 1
-        n = self.n
-        self.req[n] = evt = trio.Event()
-        try:
-            await self.send_q.put(("ask",[action,n,data]))
-            await evt.wait()
-        except BaseException:
-            self.req.pop(n)
-            raise
-        else:
-            res = self.req.pop(n)
-            if isinstance(res,Exception):
-                raise res
-            return res
+        await self._send_q.send(data)
 
 
-class Worker:
+class BaseWorker:
     """
     This is the base class for a client session. It might be interrupted
     (page reload, network glitch, or whatever).
 
-    The worker's read loop calls ``.msg_{action}`` methods with the
-    incoming message. Code in those methods must not call ``.request``
-    because that would cause a deadlock. (Don't worry, DeFramed catches
-    those.) Start a separate task with ``.spawn`` if you need to do this.
+    The talker's read loop calls :meth:`data_in` with the incoming message.
     """
     _talk = None
     _scope = None
     _nursery = None
-    _persistent_nursery = None
-    _kill_exc = None
-    _kill_flag = None
 
     title = "You forgot to set a title"
     fatal_msg = "The server had a fatal error.<br />It was logged and will be fixed soon."
@@ -222,7 +182,7 @@ class Worker:
         """
         pass
 
-    async def attach(self, talker):
+    def attach(self, talker):
         """
         Use this socket to talk. Called by the app.
 
@@ -233,18 +193,54 @@ class Worker:
         self._talk = talker
         talker.attach(self)
 
+    async def _monitor(self, *, task_status=trio.TASK_STATUS_IGNORED):
+        """
+        Background monitor. Don't override externally.
+        """
+        task_status.started()
+        pass
+
+    async def run(self, websocket):
+        """
+        Main entry point. Takes a websocket, sets up everything, then calls
+        ".talk".
+
+        You don't want to override this. Your main code should be in
+        `talk`, your setup code (called by the server) in `init`.
+        """
+        t = Talker(websocket)
+
+        try:
+            async with trio.open_nursery() as n:
+                self._nursery = n
+                await n.start(self._monitor)
+                await n.start(t.run)
+                self.attach(t)
+                with trio.CancelScope() as sc:
+                    self._scope = sc
+                    try:
+                        await self.talk()
+                    finally:
+                        self._scope = None
+        except Exception:
+            await t.died(self.fatal_msg)
+            raise
+        # not on BaseException, as in that case we close the connection and
+        # expect the client to try reconnecting
+        else:
+            await t.died(self.fatal_msg)
+
 
     async def talk(self):
         """
         Connection-specific main code. The default does nothing.
 
-        This task will get cancelled when the websocket terminates.
+        This task will get cancelled when the websocket terminates. It can
+        be used to forward external events to the web page.
         """
         pass
 
-    def cancel(self, persistent=False):
-        if persistent and self._persistent_nursery is not None:
-            self._persistent_nursery.cancel_scope.cancel()
+    def cancel(self):
         if self._scope is not None:
             self._scope.cancel()
 
@@ -252,28 +248,12 @@ class Worker:
             return
         self._scope.cancel()
 
-    async def spawn(self, task, *args, persistent=True):
+    async def spawn(self, task, *args, **kwargs):
         """
         Start a new task. Returns a cancel scope which you can use to stop
         the task.
-
-        By default, the task persists even if the client websocket
-        reconnects. Set ``persistent=False`` if you don't want that.
-
-        By default, errors get logged but don't propagate. Set ``log_exc=False``
-        if you don't want that. Or set "log_exc" to the exception, or list of
-        exceptions, you want to have logged.
         """
-        async def _spawn(task, *args, task_status=trio.TASK_STATUS_IGNORED):
-            with trio.CancelScope() as sc:
-                task_status.started(sc)
-                await task(*args)
-
-        if persistent:
-            await self._run(_spawn,task,*args)
-        else:
-            await self._nursery.start(_spawn,task,*args)
-
+        return await self._nursery.start(_spawn,task,args,kwargs)
 
     async def maybe_disconnect(self, talker):
         """internal method, called by the Talker"""
@@ -288,7 +268,82 @@ class Worker:
         """
         self.cancel()
 
-    async def getattr(self, **a: Dict[str,List[str]]) -> Dict[str,Optional[List[str]]]:
+    async def send(self, data:Any):
+        """
+        Send a message to the client.
+        """
+        await self._talk.send(data)
+
+    async def data_in(self, data):
+        """
+        Process incoming data
+        """
+        pass
+
+
+class Worker(BaseWorker):
+    """
+    This is the main class for a client session.
+
+    The talker's read loop calls ``.msg_{action}`` methods with the
+    incoming message. Code in those methods must not call ``.request``
+    because that would cause a deadlock. (Don't worry, DeFramed catches
+    those.) Start a separate task with ``.spawn`` if you need to do this.
+    """
+    _persistent_nursery = None
+    _kill_exc = None
+    _kill_flag = None
+
+    def __init__(self,*a,**k):
+        super().__init__(*a,**k)
+        self._n = 1
+        self._req = {}
+
+    async def data_in(self, data):
+        """
+        Process incoming data. Must be structured [type,data].
+
+        Replies are special:["reply",[assoc_nr,data]].
+        """
+        action,data = data
+        if action == "reply":
+            await self._reply(*data)
+            return
+        try:
+            res = getattr(self, 'msg_'+action)
+        except AttributeError:
+            res = partial(self.any_msg,action)
+        tk = processing.set((action,data))
+        try:
+            await res(data)
+        finally:
+            processing.reset(tk)
+
+    async def _reply(self, n, data):
+        if isinstance(data,Mapping) and '_error' in data:
+            data = ClientError(**data)
+        evt,self._req[n] = self._req[n],data
+        evt.set()
+
+    def cancel(self, persistent=False):
+        if persistent and self._persistent_nursery is not None:
+            self._persistent_nursery.cancel_scope.cancel()
+        super().cancel()
+
+    async def spawn(self, task, *args, persistent=True):
+        """
+        Start a new task. Returns a cancel scope which you can use to stop
+        the task.
+
+        By default, the task persists even if the client websocket
+        reconnects. Set ``persistent=False`` if you don't want that.
+        """
+        if persistent:
+            return await self._run(_spawn,task,args)
+        else:
+            return await self._nursery.start(_spawn,task,args)
+
+    async def get_attr(self, **a: Dict[str,List[str]]) -> Dict[str,Optional[List[str]]]:
         """
         Returns a list of attributes for some element.
         The element is identified by its ID.
@@ -296,14 +351,21 @@ class Worker:
         Returns: a dict with corresponding values, or None if the element
         does not exist.
         """
+        return await self.request("get_attr", a)
 
     async def exists(self, id) -> bool:
         """
         Check whether the DOM element with this ID exists.
         """
-        return (await self.getattr({id:{}}))[id] is not None
+        return (await self.get_attr({id:{}}))[id] is not None
 
-    async def msg_first(self, data) -> bool:
+    async def elem_info(self, id) -> dict:
+        """
+        Return information about this element (size, position)
+        """
+        return await self.request("elem_info", id)
+
+    async def msg_setup(self, data) -> bool:
         """
         Called when the client is connected and says Hello.
 
@@ -317,7 +379,7 @@ class Worker:
             await self.send('reload',True)
             return True
 
-        await self.send_first();
+        await self._setup();
 
         uuid = data.get('uuid')
         if uuid is None or not self.set_uuid(uuid):
@@ -368,16 +430,16 @@ class Worker:
         """
         raise UnknownActionError(name, data)
 
-    async def send_first(self):
+    async def _setup(self):
         """
         Send initial data to the client.
 
-        Called by `msg_first`.
+        Called by `msg_setup`.
         You probably should not override this.
         """
-        await self.send("first", version=self._app.version, uuid=str(self.uuid))
+        await self.send("setup", version=self._app.version, uuid=str(self.uuid))
 
-    async def send_alert(self, level, text, **kw):
+    async def alert(self, level, text, **kw):
         """
         Send a pop-up message to the client.
 
@@ -398,7 +460,30 @@ class Worker:
         """
         await self.send("info", level=level, text=text, **kw)
 
-    async def send_set(self, id: str, html: str, prepend: Optional[bool]=None):
+    async def modal_show(self, id: str = "df_modal", **opts):
+        """
+        Show a modal window.
+
+        If the ID is not given, use the built-in ``df_modal``.
+
+        Any further parameters are as described in
+        https://getbootstrap.com/docs/4.4/components/modal/.
+        """
+        if not opts:
+            opts = True
+
+        await self.send("modal",[id,opts])
+
+    async def modal_hide(self, id: str = "df_modal"):
+        """
+        Hide a modal window.
+
+        If the ID is not given, use the built-in ``df_modal``.
+        """
+        await self.send("modal",[id,False])
+
+
+    async def set_content(self, id: str, html: str, prepend: Optional[bool]=None):
         """
         Set or extend an element's innerHTML content.
 
@@ -418,9 +503,64 @@ class Worker:
             instead of (``None``) the existing content. The default is
             ``None``.
         """
-        await self.send("set", id=id, content=html, pre=prepend);
+        await self.send("set", [id, html, prepend]);
 
-    async def send_busy(self, busy: bool):
+    async def set_element(self, id: str, html: str):
+        """
+        Replace an element.
+
+        Args:
+          id:
+            the old element's ID.
+          html:
+            the element's replacement.
+        """
+        await self.send("elem", [id, html]);
+
+    async def load_style(self, id, url):
+        """
+        Load a stylesheet.
+
+        The ID is used for disambiguation.
+        """
+        await self.send("load_style", [id, url])
+
+    async def set_attr(self, id: str, **attrs):
+        """
+        Set or extend an element's attribute(s).
+        Args:
+          id:
+            the modified HTML element's ID.
+          attr=value:
+            the element's changed attributes.
+        """
+        await self.send("set_attr", [id, attrs]);
+
+    async def add_class(self, id: str, *cls):
+        """
+        Add the given classes to an element.
+
+        Args:
+          id:
+            the modified HTML element's ID.
+          cls:
+            the class to add.
+        """
+        await self.send("add_class", [id, cls]);
+
+    async def remove_class(self, id: str, *cls):
+        """
+        Remove the given classes from an element.
+
+        Args:
+          id:
+            the modified HTML element's ID.
+          cls:
+            the class(es) to remove.
+        """
+        await self.send("remove_class", [id, cls]);
+
+    async def busy(self, busy: bool):
         """
         Show/hide the client's spinner.
 
@@ -430,7 +570,7 @@ class Worker:
         """
         await self.send("busy", busy)
 
-    async def send_debug(self, val: Union[bool,str]):
+    async def debug(self, val: Union[bool,str]):
         """
         Set/clear the client's debug flag, or log a message on its console.
 
@@ -476,6 +616,14 @@ class Worker:
         """
         pass
 
+    async def msg_size(self, evt):
+        """
+        The main window's size.
+
+        This is informational.
+        """
+        pass
+
     async def any_msg(self, action: str, data):
         """
         Catch-all for unknown messages, i.e. the 'msg_{action}' handler
@@ -494,56 +642,45 @@ class Worker:
 
         You should probably call one of the specific ``send_*` methods instead.
         """
+        if action == "req":
+            raise RuntimeError("Use '.request' for that!")
         if kw:
             if data:
-                kw.update(data)
+                data.update(kw)
             else:
                 data = kw
 
-        await self._talk.send(action,data)
+        await super().send([action,data])
 
-    async def request(self, action:str, data:Any):
+    async def request(self, action:str, data:Any=None, **kw):
         """
-        Send a request to the client.
-
-        Returns the reply, or raises an error.
-
-        This method is used by `getattr` and `exists`, among others. You
-        probably should use one of these instead of calling this directly.
+        Send a request to the client, await+return the reply
         """
-        return await self._talk.request(action,data)
+        if processing.get():
+            raise RuntimeError("You cannot call this from within the receiver",processing.get())
 
-    async def run(self, websocket):
-        """
-        Main entry point. Takes a websocket, sets up everything, then calls
-        ".talk".
-
-        You don't want to override this. Your main code should be in
-        `talk`, your setup code (called by the server) in `init`.
-        """
-        t = Talker(websocket)
+        self._n += 1
+        n = self._n
+        self._req[n] = evt = trio.Event()
 
+        if kw:
+            if data:
+                data.update(kw)
+            else:
+                data = kw
         try:
-            async with trio.open_nursery() as n:
-                self._nursery = n
-                await n.start(self._kill_wait)
-                await n.start(t.run)
-                await self.attach(t)
-                with trio.CancelScope() as sc:
-                    self._scope = sc
-                    try:
-                        await self.talk()
-                    finally:
-                        self._scope = None
-        except Exception:
-            await t.died(self.fatal_msg)
+            await super().send(["req",[action,n,data]])
+            await evt.wait()
+        except BaseException:
+            self._req.pop(n)
             raise
-        # not on BaseException, as in that case we close the connection and
-        # expect the client to try reconnecting
         else:
-            await t.died(self.fatal_msg)
+            res = self._req.pop(n)
+            if isinstance(res,Exception):
+                raise res
+            return res
 
-    async def _kill_wait(self, *, task_status=trio.TASK_STATUS_IGNORED):
+    async def _monitor(self, *, task_status=trio.TASK_STATUS_IGNORED):
         self._kill_wait = trio.Event()
         task_status.started()
         await self._kill_wait.wait()
@@ -558,11 +695,11 @@ class Worker:
 
         Exceptions are propagated back to the worker.
         """
-        async def _work(proc, args, task_status=trio.TASK_STATUS_IGNORED):
+        async def _work(proc, *args, task_status=trio.TASK_STATUS_IGNORED):
             try:
                 async with trio.open_nursery() as n:
                     self._persistent_nursery = n
-                    res = await n.start(_work, proc, args)
+                    res = await n.start(proc, *args)
                     task_status.started(res)
             except Exception as exc:
                 self._kill_exc = exc
@@ -624,3 +761,18 @@ class Worker:
             return True
 
 
+class SubWorker(BaseWorker):
+    """
+    This is a worker that attaches to another via an iframe.
+    """
+    def __init__(self, worker):
+        self.master = worker
+        super().__init__(worker._app)
+        self.sub_id = self._app.attach_sub(self)
+
+    async def index(self):
+        """
+        Render the main page
+        """
+        raise RuntimeError("You need to override sending the main page")
+
